from chili.dataclasses import asdict, init_dataclass, is_dataclass
from chili.hydration import HydrationStrategy, registry, hydrate, extract
from chili.mapping import KeyMapper, Mapper

__all__ = ["asdict", "init_dataclass", "is_dataclass", "hydrate", "extract", "registry", "HydrationStrategy"]



import typing
from dataclasses import Field
from enum import Enum
from typing import Any, Dict, List, Optional, Type, Union

try:
    from typing import _TypingBase as typing_base  # type: ignore
except ImportError:
    from typing import _Final as typing_base  # type: ignore


AnnotatedTypeNames = {"AnnotatedMeta", "_AnnotatedAlias"}
_GenericAlias = getattr(typing, "_GenericAlias")
EMPTY = object()


def get_origin_type(type_name: Type) -> Optional[Type]:
    return getattr(type_name, "__origin__", None)


def get_type_args(type_name: Type) -> List[Type]:
    return getattr(type_name, "__args__", [])


def is_optional(type_name: Type) -> bool:
    return (
        get_origin_type(type_name) is Union
        and bool(get_type_args(type_name))
        and get_type_args(type_name)[-1] is type(None)
    )


def get_type_parameters(type_name: Type) -> List[Type]:
    return getattr(get_origin_type(type_name), "__parameters__", [])


def get_parameters_map(type_name: Type) -> Dict[Type, Type]:
    args = get_type_args(type_name)
    parameters = get_type_parameters(type_name)

    return dict(zip(parameters, args))


def get_dataclass_fields(type_name: Type) -> Dict[str, Field]:
    return getattr(type_name, "__dataclass_fields__", {})


def unpack_optional(type_name: Type) -> Type:
    return get_type_args(type_name)[0]


def is_enum_type(type_name: Type) -> bool:
    return issubclass(type_name, Enum)


def is_named_tuple(type_name: Type) -> bool:
    return issubclass(type_name, tuple) and hasattr(type_name, "_fields")


def is_typed_dict(type_name: Type) -> bool:
    return issubclass(type_name, dict) and hasattr(type_name, "__annotations__")


def map_generic_type(type_name: Any, type_map: Dict[Any, Any]) -> Any:
    if not type_map:
        return type_name

    type_args = get_type_args(type_name)
    origin_type = get_origin_type(type_name)

    if origin_type is None:  # not a generic type
        return type_name

    if type_args:
        mapped_args = tuple([type_map[arg] for arg in type_args])
        return _GenericAlias(origin_type, mapped_args)

    return type_name



from dataclasses import is_dataclass
from typing import Any, Dict, Type, TypeVar

from .hydration import hydrate, extract
from .mapping import Mapper

T = TypeVar("T")


def init_dataclass(data: Dict[str, Any], dataclass: Type[T], mapping: Mapper = None) -> T:
    return hydrate(data, dataclass, strict=False, mapping=mapping)


def asdict(data: Any, mapping: Mapper = None) -> Dict[str, Any]:
    if not is_dataclass(data):
        raise TypeError(f"Can extract only dataclasses, passed `{type(data)}` type instead.")

    return extract(data, strict=False, mapping=mapping)



from collections import namedtuple
from collections.abc import Iterable
from typing import Union, Dict, Callable, Any

KeyMapper = namedtuple("KeyMapper", "key scheme")
MappingScheme = Dict[str, Union[Dict, str, Callable]]


class Mapper:
    def __init__(self, scheme: MappingScheme):
        self.scheme = scheme

    def map(self, data) -> Any:
        return self._map(data, self.scheme)

    def _map(self, data: Dict, scheme: MappingScheme) -> Any:
        result = {}
        for new_key, old_key in scheme.items():
            if isinstance(old_key, int):
                if old_key:
                    result[new_key] = data.get(new_key, None)
                continue
            elif isinstance(old_key, str):
                result[new_key] = data.get(old_key, None)
                continue
            elif isinstance(old_key, KeyMapper):
                if old_key.key not in data:
                    result[new_key] = None
                value = data[old_key.key]
                item_scheme = old_key.scheme
            elif isinstance(old_key, dict):
                if new_key not in data:
                    result[new_key] = None
                    continue
                value = data[new_key]
                item_scheme = old_key
            elif callable(old_key):
                result[new_key] = old_key(data)
                continue
            else:
                raise ValueError(
                    f"Unexpected mapping schema passed in `{new_key}`, "
                    f"expected string, callable, bool or instance of KeyMapper."
                )

            if callable(item_scheme):
                result[new_key] = item_scheme(value)
                continue
            elif isinstance(value, dict):
                result[new_key] = self._map(value, item_scheme)
                continue
            elif isinstance(value, Iterable):
                result[new_key] = [self._map(item, item_scheme) for item in iter(value)]
                continue
            else:
                raise ValueError(f"Could not map value {value} with provided scheme {item_scheme}")

        return result



import collections
import datetime
from abc import abstractmethod
from dataclasses import MISSING, is_dataclass
from decimal import Decimal
from enum import Enum
from functools import partial
from inspect import isclass
from typing import (
    Any,
    AnyStr,
    Callable,
    Deque,
    Dict,
    FrozenSet,
    Generic,
    List,
    NamedTuple,
    Sequence,
    Set,
    Tuple,
    Type,
    TypeVar,
    Union,
)

from typing_extensions import Protocol, TypedDict

from .iso_datetime import (
    parse_iso_date,
    parse_iso_datetime,
    parse_iso_duration,
    parse_iso_time,
    timedelta_to_iso_duration,
)
from .mapping import Mapper, MappingScheme
from .typing import (
    get_dataclass_fields,
    get_origin_type,
    get_parameters_map,
    get_type_args,
    is_enum_type,
    is_named_tuple,
    is_optional,
    is_typed_dict,
    map_generic_type,
    unpack_optional,
)

T = TypeVar("T")


class HydrationStrategy(Protocol):
    @abstractmethod
    def hydrate(self, value: Any) -> Any:
        ...

    @abstractmethod
    def extract(self, value: Any) -> Any:
        ...


class DataclassStrategy(HydrationStrategy):
    def __init__(self, dataclass_name: Type):
        self._getters: Dict[str, Callable] = {}
        self._setters: Dict[str, Callable] = {}
        self._dataclass_name = dataclass_name
        self._set_strategies()

    def hydrate(self, value: Any) -> Any:
        instance = self._dataclass_name.__new__(self._dataclass_name)  # type: ignore

        for setter in self._setters.values():
            setter(instance, value)

        if hasattr(instance, "__post_init__"):
            instance.__post_init__()

        return instance

    def extract(self, value: Any) -> Any:
        if not isinstance(value, self._dataclass_name):
            raise ValueError(f"Passed value must be type of {self._dataclass_name}.")

        result = {}
        for name, get in self._getters.items():
            result[name] = get(getattr(value, name, None))

        return result

    def _get_strategy(self, field_type: Type) -> HydrationStrategy:
        return registry.get_for(field_type)

    def _set_strategies(self) -> None:
        fields = get_dataclass_fields(self._dataclass_name)
        for field_name, field_descriptor in fields.items():
            if not field_descriptor.init and not field_descriptor.repr:
                continue

            strategy = self._get_strategy(field_descriptor.type)
            if field_descriptor.init:
                self._setters[field_name] = partial(
                    set_dataclass_property,
                    setter=strategy.hydrate,
                    property_name=field_name,
                    default_factory=field_descriptor.default_factory,  # type: ignore
                    default_value=field_descriptor.default,
                    optional=is_optional(field_descriptor.type),
                )

            if field_descriptor.repr:
                self._getters[field_name] = strategy.extract


class GenericDataclassStrategy(DataclassStrategy):
    def __init__(self, dataclass_name: Type):
        self._generic_type = dataclass_name
        self._generic_parameters = get_parameters_map(dataclass_name)
        type_: Type = get_origin_type(dataclass_name)  # type: ignore
        super().__init__(type_)

    def _get_strategy(self, field_type: Type) -> HydrationStrategy:
        return registry.get_for(map_generic_type(field_type, self._generic_parameters))


class DummyStrategy(HydrationStrategy):
    def hydrate(self, value: Any, *args) -> Any:
        return value

    def extract(self, value: Any, *args) -> Any:
        return value


class SimpleStrategy(HydrationStrategy):
    def __init__(self, hydrate_type: Type, extract_type: Type):
        self._hydrate_type = hydrate_type
        self._extract_type = extract_type

    def hydrate(self, value: Any) -> Any:
        return self._hydrate_type(value)

    def extract(self, value: Any) -> Any:
        return self._extract_type(value)


class ListStrategy(HydrationStrategy):
    def __init__(self, subtype: HydrationStrategy):
        self._subtype = subtype

    def hydrate(self, value: Any) -> Any:
        return [self._subtype.hydrate(item) for item in value]

    def extract(self, value: Any) -> Any:
        return [self._subtype.extract(item) for item in value]


class SetStrategy(ListStrategy):
    def hydrate(self, value: Any) -> Any:
        return set(super().hydrate(value))


class FrozenSetStrategy(ListStrategy):
    def hydrate(self, value: Any) -> Any:
        return frozenset(super().hydrate(value))


class DequeStrategy(ListStrategy):
    def hydrate(self, value: Any) -> Any:
        return collections.deque(super().hydrate(value))


class TupleStrategy(HydrationStrategy):
    def __init__(self, *subtypes: HydrationStrategy):
        self._subtypes = subtypes
        self._subtypes_length = len(subtypes)
        self._is_ellipsis_present = self._subtypes[-1] is ...

    def hydrate(self, value: Any) -> Any:
        if self._is_ellipsis_present:
            return self._hydrate_ellipsis_tuple(value)

        result = []
        for index, subtype in enumerate(self._subtypes):
            result.append(subtype.hydrate(value[index]))

        return tuple(result)

    def extract(self, value: Any) -> Any:
        if self._is_ellipsis_present:
            return self._extract_ellipsis_tuple(value)

        return [subtype.extract(value[index]) for index, subtype in enumerate(self._subtypes)]

    def _hydrate_ellipsis_tuple(self, value) -> Any:
        last_known_strategy = self._subtypes[0]
        result = []
        for index, item in enumerate(value):
            if index + 1 < self._subtypes_length:
                last_known_strategy = self._subtypes[index]

            result.append(last_known_strategy.hydrate(item))

        return tuple(result)

    def _extract_ellipsis_tuple(self, value) -> Any:
        last_known_strategy = self._subtypes[0]
        result = []
        for index, item in enumerate(value):
            if index + 1 < self._subtypes_length:
                last_known_strategy = self._subtypes[index]

            result.append(last_known_strategy.extract(item))

        return result


class NamedTupleStrategy(HydrationStrategy):
    def __init__(self, class_name: Type[NamedTuple]):
        self._class_name = class_name
        self._is_typed = hasattr(class_name, "__annotations__")
        self._arg_strategies: List[HydrationStrategy] = []
        if self._is_typed:
            self._build_type_mapper(class_name.__annotations__)

    def hydrate(self, value: Any) -> Any:
        if not self._is_typed:
            return self._class_name(*value)

        hydrated_values = []
        for index, item in enumerate(value):
            if index < len(self._arg_strategies):
                hydrated_values.append(self._arg_strategies[index].hydrate(item))
                continue
            hydrated_values.append(item)

        return self._class_name(*hydrated_values)

    def extract(self, value: Any) -> Any:
        result = list(value)
        if not self._is_typed:
            return result
        extracted_values = []
        for index, item in enumerate(result):
            if index < len(self._arg_strategies):
                extracted_values.append(self._arg_strategies[index].extract(item))
                continue
            extracted_values.append(item)

        return extracted_values

    def _build_type_mapper(self, field_types: Dict[str, Type]) -> None:
        for item_type in field_types.values():
            self._arg_strategies.append(registry.get_for(item_type))


class EnumStrategy(HydrationStrategy):
    def __init__(self, class_name: Type[Enum]):
        self._class_name = class_name

    def hydrate(self, value: Any) -> Any:
        return self._class_name(value)

    def extract(self, value: Any) -> Any:
        return value.value


class DictStrategy(HydrationStrategy):
    def __init__(self, key: HydrationStrategy, value: HydrationStrategy):
        self._key = key
        self._value = value

    def hydrate(self, value: Any) -> Any:
        return {self._key.hydrate(key): self._value.hydrate(item) for key, item in value.items()}

    def extract(self, value: Any) -> Any:
        return {self._key.extract(key): self._value.extract(item) for key, item in value.items()}


class OrderedDictStrategy(DictStrategy):
    def hydrate(self, value: Any) -> Any:
        return collections.OrderedDict(super().hydrate(value))


class TypedDictStrategy(HydrationStrategy):
    def __init__(self, type_name: Type):
        self._strategies: Dict[str, HydrationStrategy] = {}
        for key_name, key_type in type_name.__annotations__.items():
            self._strategies[key_name] = registry.get_for(key_type)

    def hydrate(self, value: Any) -> Any:
        return {key: self._strategies[key].hydrate(item) for key, item in value.items()}

    def extract(self, value: Any) -> Any:
        return {key: self._strategies[key].extract(item) for key, item in value.items()}


class DateStrategy(HydrationStrategy):
    """
    Conforms ISO 8601 standard https://www.iso.org/iso-8601-date-and-time-format.html
    """

    def hydrate(self, value: Any) -> Any:
        if isinstance(value, datetime.date):
            return value

        return parse_iso_date(value)

    def extract(self, value: Any) -> Any:
        return value.isoformat()


class DateTimeStrategy(HydrationStrategy):
    """
    Conforms ISO 8601 standard https://www.iso.org/iso-8601-date-and-time-format.html
    """

    def hydrate(self, value: Any) -> Any:
        if isinstance(value, datetime.datetime):
            return value

        return parse_iso_datetime(value)

    def extract(self, value: Any) -> Any:
        return value.isoformat()


class TimeStrategy(HydrationStrategy):
    """
    Conforms ISO 8601 standard https://www.iso.org/iso-8601-date-and-time-format.html
    """

    def hydrate(self, value: Any) -> Any:
        if isinstance(value, datetime.time):
            return value

        return parse_iso_time(value)

    def extract(self, value: Any) -> Any:
        return value.isoformat()


class TimeDeltaStrategy(HydrationStrategy):
    """
    Conforms ISO 8601 standard https://www.iso.org/iso-8601-date-and-time-format.html
    """

    def hydrate(self, value: Any) -> Any:
        if isinstance(value, datetime.timedelta):
            return value

        return parse_iso_duration(value)

    def extract(self, value: Any) -> Any:
        return timedelta_to_iso_duration(value)


class OptionalTypeStrategy(HydrationStrategy):
    def __init__(self, type_strategy: HydrationStrategy):
        self._type_strategy = type_strategy

    def hydrate(self, value: Any, *args) -> Any:
        if value is None:
            return None

        return self._type_strategy.hydrate(value)

    def extract(self, value: Any, *args) -> Any:
        if value is None:
            return None

        return self._type_strategy.extract(value)


class UnionStrategy(HydrationStrategy):
    PRIMITIVE_TYPES = {int, float, bool, str}

    def __init__(self, types: List[Type]):
        self.supported_types = types
        self.primitive_types = []
        self.dataclass_types = []
        self.other_types = []

        for i_type in types:
            if i_type in UnionStrategy.PRIMITIVE_TYPES:
                self.primitive_types.append(i_type)
                continue
            if is_dataclass(i_type):
                self.dataclass_types.append(i_type)
                continue

            self.other_types.append(i_type)

    def hydrate(self, value: Any) -> Any:
        value_type = type(value)

        # primitives
        if value_type in UnionStrategy.PRIMITIVE_TYPES:
            if value_type in self.primitive_types:
                return registry.get_for(value_type).hydrate(value)

            for try_type in self.primitive_types:
                try:
                    return registry.get_for(try_type).hydrate(value)
                except Exception:
                    continue

            raise ValueError("Could not hydrate passed value.")

        # dataclasses
        if value_type is dict:
            for dataclass in self.dataclass_types:
                try:
                    dataclass(**value)
                    return registry.get_for(dataclass).hydrate(value)
                except Exception:
                    continue

        # other types
        for try_type in self.other_types:
            try:
                return registry.get_for(try_type).hydrate(value)
            except Exception:
                continue

        raise ValueError("Could not hydrate passed value.")

    def extract(self, value: Any) -> Any:
        value_type = type(value)
        if value_type not in self.supported_types:
            raise ValueError("Could not extract passed value.")

        return registry.get_for(value_type).extract(value)


def set_dataclass_property(
    obj: object,
    attributes: Dict[str, Any],
    property_name: str,
    setter: Callable,
    default_factory: Union[Callable, Any],
    default_value: Any,
    optional: bool,
) -> None:
    if property_name in attributes:
        setattr(obj, property_name, setter(attributes[property_name]))
        return

    if callable(default_factory):
        setattr(obj, property_name, default_factory())
        return

    if default_value is not MISSING:
        setattr(obj, property_name, default_value)
        return

    attribute_value = attributes.get(property_name, MISSING)

    if attribute_value is MISSING:
        if optional:
            setattr(obj, property_name, None)
            return
        raise AttributeError(f"Property `{property_name}` is required.")

    try:
        setattr(obj, property_name, setter(attribute_value))
    except Exception as error:
        raise AttributeError(f"Could not hydrate `{property_name}` property with `{attribute_value}` value.") from error


StrategyDictionary = Dict[Any, HydrationStrategy]


class StrategyRegistry:
    _cached: StrategyDictionary = {}
    _custom: StrategyDictionary = {}
    _builtin_types: StrategyDictionary = {
        bool: SimpleStrategy(bool, bool),
        collections.OrderedDict: SimpleStrategy(collections.OrderedDict, dict),
        int: SimpleStrategy(int, int),
        float: SimpleStrategy(float, float),
        str: SimpleStrategy(str, str),
        bytes: SimpleStrategy(bytes, bytes),
        list: SimpleStrategy(list, list),
        set: SimpleStrategy(set, list),
        frozenset: SimpleStrategy(frozenset, list),
        tuple: SimpleStrategy(tuple, list),
        dict: SimpleStrategy(dict, dict),
        datetime.time: TimeStrategy(),
        datetime.date: DateStrategy(),
        datetime.datetime: DateTimeStrategy(),
        datetime.timedelta: TimeDeltaStrategy(),
        collections.deque: SimpleStrategy(collections.deque, list),
        TypedDict: SimpleStrategy(dict, dict),  # type: ignore
        Dict: SimpleStrategy(dict, dict),  # type: ignore
        List: SimpleStrategy(list, list),
        Sequence: SimpleStrategy(list, list),
        Tuple: SimpleStrategy(tuple, list),  # type: ignore
        Set: SimpleStrategy(set, list),
        FrozenSet: SimpleStrategy(frozenset, list),
        Deque: SimpleStrategy(collections.deque, list),
        AnyStr: SimpleStrategy(str, str),
        Any: DummyStrategy(),  # type: ignore
        Decimal: SimpleStrategy(Decimal, str),
    }
    _builtin_generics: Dict[Any, Type] = {
        list: ListStrategy,
        tuple: TupleStrategy,
        dict: DictStrategy,
        set: SetStrategy,
        frozenset: FrozenSetStrategy,
        collections.deque: DequeStrategy,
        collections.OrderedDict: OrderedDictStrategy,
        Union: UnionStrategy,
    }

    def add(self, type_name: Type, strategy: HydrationStrategy):
        self._custom[type_name] = strategy

    def get_for(self, type_name: Type, strict: bool = False) -> HydrationStrategy:
        # Customized type
        if type_name in self._custom:
            return self._custom[type_name]

        # Simple types
        if type_name in self._builtin_types:
            return self._builtin_types[type_name]

        # Already resolved types
        if type_name in self._cached:
            return self._cached[type_name]

        # Dataclasses
        if is_dataclass(type_name):
            if issubclass(type_name, Generic):  # type: ignore
                raise ValueError("Cannot automatically hydrate/extract non-parametrised generic classes.")

            self._cached[type_name] = DataclassStrategy(type_name)
            return self._cached[type_name]

        origin_type = get_origin_type(type_name)

        # Other non-generic types
        if origin_type is None:
            if not isclass(type_name):
                if strict:
                    raise ValueError(f"Cannot hydrate/extract type `{type_name}`")
                return self._builtin_types[Any]  # type: ignore

            if is_enum_type(type_name):
                self._cached[type_name] = EnumStrategy(type_name)
                return self._cached[type_name]

            if is_named_tuple(type_name):
                self._cached[type_name] = NamedTupleStrategy(type_name)
                return self._cached[type_name]

            if is_typed_dict(type_name):
                self._cached[type_name] = TypedDictStrategy(type_name)
                return self._cached[type_name]

            if strict:
                raise ValueError(f"Cannot hydrate/extract type `{type_name}`")

            return self._builtin_types[Any]  # type: ignore

        # Generic dataclass
        if is_dataclass(origin_type):
            self._cached[type_name] = GenericDataclassStrategy(type_name)
            return self._cached[type_name]

        if origin_type not in self._builtin_generics:
            # Unknown type, just ignore it
            if not is_optional(type_name):
                if strict:
                    raise ValueError(f"Cannot hydrate/extract type `{type_name}`")
                return self._builtin_types[Any]  # type: ignore

            # Optional type
            self._cached[type_name] = OptionalTypeStrategy(self.get_for(unpack_optional(type_name)))
            return self._cached[type_name]

        if origin_type is Union:
            type_args = get_type_args(type_name)
            if len(type_args) == 2 and type_args[-1] is type(None):
                self._cached[type_name] = OptionalTypeStrategy(self.get_for(type_args[0]))
            else:
                self._cached[type_name] = UnionStrategy(get_type_args(type_name))
            return self._cached[type_name]

        # Resolve generic attributes
        type_attributes: List[Union[HydrationStrategy, Any]] = []
        for subtype in get_type_args(type_name):
            if subtype is ...:
                type_attributes.append(...)
                continue
            type_attributes.append(self.get_for(subtype))

        # Generic types
        self._cached[type_name] = self._builtin_generics[origin_type](*type_attributes)
        return self._cached[type_name]


registry = StrategyRegistry()


def hydrate(data: Any, type_name: Type[T], strict: bool = True, mapping: Mapper = None) -> T:
    if mapping is not None:
        data = mapping.map(data)

    strategy = registry.get_for(type_name, strict)
    return strategy.hydrate(data)


def extract(data: Any, strict: bool = True, mapping: Mapper = None) -> Any:
    strategy = registry.get_for((type(data)), strict)

    if mapping is not None:
        return mapping.map(strategy.extract(data))

    return strategy.extract(data)


__all__ = ["registry", "HydrationStrategy", "hydrate", "extract"]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['chili']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'chili',
    'version': '1.3.1',
    'description': 'Chili is a dataclass support library. It is providing simple and fast hydration and extraction interfaces for dataclasses.',
    'long_description': '# Chili [![codecov](https://codecov.io/gh/kodemore/chili/branch/main/graph/badge.svg?token=TCG7SRQFD5)](https://codecov.io/gh/kodemore/chili) [![CI](https://github.com/kodemore/chili/actions/workflows/main.yaml/badge.svg?branch=main)](https://github.com/kodemore/chili/actions/workflows/main.yaml) [![Release](https://github.com/kodemore/chili/actions/workflows/release.yml/badge.svg)](https://github.com/kodemore/chili/actions/workflows/release.yml) [![License: MIT](https://img.shields.io/badge/License-MIT-yellow.svg)](https://opensource.org/licenses/MIT)\nChili is an extensible dataclass support library.\nIt contains helper functions to simplify initialising and extracting complex dataclasses.\nThis might come in handy when you want to transform your request\'s data to well-defined and easy to understand objects \nor when there is a need to hydrate and/or transform database records to desired representation in memory.\n\nLibrary also ensures type integrity and provides simple interface, which does not pollute your codebase\nwith unwanted abstractions.\n\n## Features\n- extensible and easy to use\n- initialisation and extraction for complex dataclasses\n- supports most python\'s types found in `typing` package (including generics)\n- does not pollute your codebase\n- supports adding custom types\n- data mapping/transformation with `chili.Mapping`\n\n## Installation\n\nWith pip,\n```shell\npip install chili\n```\nor through poetry\n```shell\npoetry add chili\n```\n\n# Usage\n\n## Initialising a dataclass\n\n```python\nfrom dataclasses import dataclass\nfrom typing import List\nfrom chili import init_dataclass\n\n@dataclass\nclass Tag:\n    id: str\n    name: str\n\n@dataclass\nclass Pet:\n    name: str\n    tags: List[Tag]\n    age: int\n\npet = init_dataclass({"name": "Bobik", "tags": [{"name": "dog", "id": "12"}]}, Pet)\nassert isinstance(pet, Pet)\nassert isinstance(pet.tags, List)\nassert isinstance(pet.tags[0], Tag)\n```\n\n> This example shows how simply cast your dict to given dataclass and that type are ensured by the init_dataclass function.\n\n## Transforming dataclass back to a dict\n\n```python\nfrom chili import asdict\nfrom dataclasses import dataclass\n\n@dataclass\nclass Money:\n    currency: str\n    amount: float\n\nsome_pounds = Money("GBP", "100.00")\n\nsome_pounds_dict = asdict(some_pounds)\n\nassert isinstance(some_pounds_dict, dict)\nassert isinstance(some_pounds_dict["amount"], float)\n```\n\n> Chili works with wide commonly used python types, but not every type can be simply transformed back and forth, \n> so make sure you familiarise yourself with supported types.\n\n## Using default values\n\n```python\nfrom dataclasses import dataclass, field\nfrom typing import List\nfrom chili import init_dataclass\n\n\n@dataclass\nclass Pet:\n    name: str\n    tags: List[str] = field(default_factory=lambda: ["pet"])\n\n\nboo = init_dataclass({"name": "Boo"}, Pet)\n\nassert isinstance(boo, Pet)\nassert boo.tags == [\'pet\']\n```\n\nIn the above example tags attribute was not available in the dict object, so default value set \nin dataclass is being used instead.\n\n\n> Please note `dataclasses` module does not allow mutable values in `default` argument of the `field` function, \n> so this example is using `default_factory` instead. More details about dataclasses\' `default` and `default_factory` \n> arguments are available in the [python\'s documentation](https://docs.python.org/3/library/dataclasses.html#dataclasses.field).\n\n## Hiding fields from hydration/deserialisation\n\nThere might be scenarios where not all dataclass\' fields should be hydrated. In this scenario use built-in python\'s `dataclasses` module `field` function, like in the example below:\n\n```python\nfrom chili import init_dataclass\nfrom dataclasses import dataclass, field\nfrom typing import List\n\n\n@dataclass\nclass Pet:\n    name: str\n    tags: List[str]\n    tags_length: int = field(init=False)\n\n    def __post_init__(self):\n        self.tags_length = len(self.tags)\n\n\nboo = init_dataclass({"name": "Boo", "tags": ["hamster", "boo"], "tags_length": 0}, Pet)\n\nassert isinstance(boo, Pet)\nassert boo.tags_length == 2\n```\n\nIn the above example length of the `tags` attribute is recalculated everytime we initialise the class \nand hydrating it might be superfluous.\n\n\n## Hiding fields from extraction/serialisation\n\nTo hide attributes of dataclass from being extracted into dict simply use `field` function with `repr` attribute set to `False`\n\n```python\nfrom dataclasses import dataclass, field\nfrom typing import List\n\nfrom chili import asdict\n\n\n@dataclass\nclass Pet:\n    name: str\n    tags: List[str] = field(repr=False)\n\n\nboo = Pet(name="Boo", tags=["pet", "hamster", "powerful!"])\n\nboo_dict = asdict(boo)\n\nassert "tags" not in boo_dict\n```\n\n## Data mapping\n\nSometimes you might run into scenarios that data coming from different sources needs to be remapped \nbefore you can hydrate it to your dataclass. There might be several reasons for that:\n- input data is using camelCase convention\n- input data is using different naming\n- input data is missing values\n\nIn all those cases you can pass `mapping` attribute to `init_dataclass/hydrate` or `asdict/extract` functions to perform\nmapping before hydration or after extraction dataclass.\n\n### Simple mapping\nPlease consider the following example of simple name mapping:\n\n```python\nfrom dataclasses import dataclass\nfrom typing import List\n\nimport chili\n\ninput_data = {\n    "petName": "Bobik",\n    "age": "12",\n    "taggedWith": [\n        {"tagName": "smart"},\n        {"tagName": "dog"},\n        {"tagName": "happy"},\n    ]\n}\n\n\n@dataclass\nclass Pet:\n    name: str\n    age: int\n    tags: List[dict]\n\n\nmapping = chili.Mapper({\n    "name": "petName",  # `petName` will be renamed to `name`, which corresponds to `Pet.name` field\n    "age": True,  # we just pass true value to include field "as is"\n    "tags": chili.KeyMapper("taggedWith", {  # `taggedWith` is a complex structure we want to map, so we have to use KeyMapper \n        "name": "tagName",  # `tagName` will be renamed to `name` which corresponds to `Pet.tags[{index}].name`\n    }),\n})\n\nbobik = chili.hydrate(input_data, Pet, mapping=mapping)\nprint(bobik)  # Pet(name=\'Bobik\', age=12, tags=[{\'name\': \'smart\'}, {\'name\': \'dog\'}, {\'name\': \'happy\'}])\n```\n\n### Mappings with custom behaviour\n\nWe can also use lambdas and functions in mapping to achieve the same result as in the previous example.\n\n```python\nfrom dataclasses import dataclass\nfrom typing import List, Tuple\n\nimport chili\n\n\ndef map_pet_tags(value: List) -> List:\n    return [{"name": item["tagName"]} for item in value]\n\n\ninput_data = {\n    "petName": "Bobik",\n    "petAge": "12",\n    "taggedWith": [\n        {"tagName": "smart"},\n        {"tagName": "dog"},\n        {"tagName": "happy"},\n    ]\n}\n\n\n@dataclass\nclass Pet:\n    name: str\n    age: int\n    tags: List[dict]\n\n\nmapping = chili.Mapper({\n    "name": "petName",\n    "age": lambda value: value["petAge"],  # callables will always receive current\'s scope data as input \n    "tags": chili.KeyMapper("taggedWith", map_pet_tags)\n})\n\nbobik = chili.hydrate(input_data, Pet, mapping=mapping)\nprint(bobik)  # Pet(name=\'Bobik\', age=12, tags=[{\'name\': \'smart\'}, {\'name\': \'dog\'}, {\'name\': \'happy\'}])\n```\n\n## Declaring custom hydrators\n\nIf you work with types that are neither dataclasses nor directly supported by `Chili`, you can define your own\nhydrator to customise how the type is initialised and how it should be de-initialised by declaring a subclass of\n`chili.hydration.HydrationStrategy` and registering it, like below:\n\n```python\nfrom chili import hydrate, registry, extract, HydrationStrategy\nimport  typing\n\nclass MyType:\n    def __init__(self, value):\n        self.value = value\n\nclass MyHydrator(HydrationStrategy):\n    def extract(self, value):  # value will be instance of MyType\n        return value.value\n        \n    def hydrate(self, value):\n        return MyType(value)\n\n# register our custom type in the hydration registry    \nregistry.add(MyType, MyHydrator())\n\n# usage\nassert isinstance(hydrate("hello", MyType), MyType)\nassert isinstance(hydrate("hello", typing.Optional[MyType]), MyType) # this will work as well with optional types\n\nassert extract(MyType("hello")) == "hello"\n```\n\n\n## Working with `Generic` types\n\n`Chili` support most of python\'s generic types like; `typing.List`, `typing.Tuple`, `typing.Dict`, etc. \nSupport is also provided for generic types defined by user (to some extent).\n\n```python\nfrom dataclasses import dataclass\nfrom typing import Generic, List, TypeVar\nfrom chili import init_dataclass\n\nT = TypeVar("T")\n\n@dataclass\nclass Pet:\n    name: str\n\n@dataclass\nclass Animal:\n    name: str\n\n@dataclass\nclass CustomList(Generic[T]):\n    list: List[T]\n\n\npet_list = init_dataclass(\n    {"list": [\n        {"name": "Boo"},\n        {"name": "Bobek"},\n    ]},\n    CustomList[Pet]\n)\n\nassert isinstance(pet_list, CustomList)\nfor pet in pet_list.list:\n    assert isinstance(pet, Pet)\n\n\nanimal_list = init_dataclass(\n    {"list": [\n        {"name": "Boo"},\n        {"name": "Bobek"},\n    ]},\n    CustomList[Animal]\n)\n\nassert isinstance(pet_list, CustomList)\nfor animal in animal_list.list:\n    assert isinstance(animal, Animal)\n```\n\nIn the above example there are three definitions of dataclasses: `Pet`, `Animal` and `CustomList`. \n`Pet` and `Animal` are just ordinary dataclasses but `CustomList` is a generic class, parametrised with `T` parameter. \nThis means we can have subtypes, like: `CustomList[Pet]`, `CustomList[Animal]` or even `CustomList[Dict]`.\n\n`init_dataclass` function understands that passed type is a generic type, and can handle it as suspected. \n\nHydration of dataclass inheriting from another generic dataclasses is also supported, \nonly if that dataclass specifies the parameters:\n\n```python\nfrom dataclasses import dataclass\nfrom typing import Generic, List, TypeVar\nfrom chili import init_dataclass\n\nT = TypeVar("T")\n\n@dataclass\nclass Pet:\n    name: str\n\n@dataclass\nclass Animal:\n    name: str\n\n@dataclass\nclass CustomList(Generic[T]):\n    list: List[T]\n\n@dataclass\nclass ExtendedGenericList(CustomList, Generic[T]):\n    ...\n\n@dataclass\nclass ExtendedList(CustomList[Pet]):\n    ...\n\n# this will work\npet_list = init_dataclass(\n    {"list": [\n        {"name": "Boo"},\n        {"name": "Bobek"},\n    ]},\n    ExtendedGenericList[Pet]\n)\n\n# this will fail\nfailed_pet_list = init_dataclass(\n    {"list": [\n        {"name": "Boo"},\n        {"name": "Bobek"},\n    ]},\n    ExtendedList\n)\n```\n\nIn the above example `ExtendedList` will fail during initialisation, the reason for that is information \nrequired to parametrise this class and probably its subclasses or any other classes aggregated by this class is lost. \nFor now this behaviour is not supported for auto-hydration mode. `ExtendedGenericList[Pet]` will work as expected.\n\n## Supported data types\n\n### `bool`\n\nPassed value is automatically hydrated to boolean with python\'s built-in `bool` on hydration and extraction.\n\n### `dict`\n\nPassed value is automatically hydrated to dict with python\'s built-in `dict` on hydration and extraction.\n\n### `float`\n\nPassed value is automatically hydrated to float with python\'s built-in `float` on hydration and extraction.\n\n### `frozenset`\n\nPassed value is automatically hydrated to frozen set with python\'s built-in `frozenset` and extracted to `list`.\n\n### `int`\n\nPassed value is automatically hydrated to int with python\'s built-in `int` on hydration and extraction.\n\n### `list`\n\nPassed value is automatically hydrated to list with python\'s built-in `list` on hydration and extraction.\n\n### `set`\n\nPassed value is automatically hydrated to set with python\'s built-in `set` and extracted to `list`.\n\n### `str`\n\nPassed value is automatically hydrated to string with python\'s built-in `str` on hydration and extraction.\n\n### `tuple`\n\nPassed value is automatically hydrated to tuple with python\'s built-in `tuple` and extracted to `list`.\n\n### `collections.namedtuple`\n\nPassed value is automatically hydrated to named tuple and extracted to `list`.\n\n### `collections.deque`\n\nPassed value is automatically hydrated to an instance of `collections.deque` and extracted to `list`.\n\n### `collections.OrderedDict`\n\nPassed value is automatically hydrated to an instance of `collections.OrderedDict` and extracted to `dict`.\n\n### `datetime.date`\n\nPassed value must be valid ISO-8601 date string, then it is automatically hydrated to an instance of `datetime.date` \nclass and extracted to ISO-8601 format compatible string.\n\n### `datetime.datetime`\n\nPassed value must be valid ISO-8601 date time string, then it is automatically hydrated to an instance of `datetime.datetime` \nclass and extracted to ISO-8601 format compatible string.\n\n### `datetime.time`\n\nPassed value must be valid ISO-8601 time string, then it is automatically hydrated to an instance of `datetime.time` \nclass and extracted to ISO-8601 format compatible string.\n\n### `datetime.timedelta`\n\nPassed value must be valid ISO-8601 duration string, then it is automatically hydrated to an instance of `datetime.timedelta`\nclass and extracted to ISO-8601 format compatible string.\n\n### `decimal.Decimal`\n\nPassed value must be a string containing valid decimal number representation, for more please read python\'s manual\nabout [`decimal.Decimal`](https://docs.python.org/3/library/decimal.html#decimal.Decimal), on extraction value is\nextracted back to string.\n\n### `enum.Enum`\n\nSupports hydration of all instances of `enum.Enum` subclasses as long as value can be assigned\nto one of the members defined in the specified `enum.Enum` subclass. During extraction the value is\nextracted to value of the enum member.\n\n### `enum.IntEnum`\n\nSame as `enum.Enum`.\n\n### `typing.Any`\n\nPassed value is unchanged during hydration and extraction process.\n\n### `typing.AnyStr`\n\nSame as `str`\n\n### `typing.Deque`\n\nSame as `collection.dequeue` with one exception, if subtype is defined, eg `typing.Deque[int]` each item inside queue\nis hydrated accordingly to subtype.\n\n### `typing.Dict`\n\nSame as `dict` with exception that keys and values are respectively hydrated and extracted to match\nannotated type.\n\n### `typing.FrozenSet`\n\nSame as `frozenset` with exception that values of a frozen set are respectively hydrated and extracted to\nmatch annotated type.\n\n### `typing.List`\n\nSame as `list` with exception that values of a list are respectively hydrated and extracted to match annotated type.\n\n### `typing.NamedTuple`\n\nSame as `namedtuple`.\n\n### `typing.Optional`\n\nOptional types can carry additional `None` value which chili\'s hydration process will respect, so for example \nif your type is `typing.Optional[int]` `None` value is not hydrated to `int`.\n\n### `typing.Set`\n\nSame as `set` with exception that values of a set are respectively hydrated and extracted to match annotated type.\n\n### `typing.Tuple`\n\nSame as `tuple` with exception that values of a set are respectively hydrated and extracted to match annotated types.\nEllipsis operator (`...`) is also supported.\n\n### `typing.TypedDict`\n\nSame as `dict` but values of a dict are respectively hydrated and extracted to match annotated types. \n\n\n### `typing.Generic\n\nOnly parametrised generic classes are supported, dataclasses that extends other Generic classes without parametrisation will fail.\n\n\n### `typing.Union`\n\nLimited support for Unions.\n\n## API\n\n### **`chili.hydrate`**(**`value`**: _`typing.Any`_, **`type_name`**: _`Type[T]`_, **`strict`**: _`bool`_ = `False`, **`mapping`**: _`chili.Mapper`_ = `None`) -> _`T`_\n\nHydrates given value into instance of passed type. If hydration fails, it returns passed value as a result, \nif strict mode is set to `True` it raises `InvalidValueError`.\n\n### **`chili.extract`**(**`value`**: _`typing.Any`_, **`strict`**: _`bool`_ = `False`, **`mapping`**: _`chili.Mapper`_ = `None`) -> _`typing.Any`_\n\nExtracts given value into primitive or set of primitives. If extraction fails, it returns passed value as a result, if\nstrict mode is set to `True` it raises `InvalidValueError`.\n\n### **`chili.init_dataclass`**(**`value`**: _`dict`_, **`type_name`**: _`Type[T]`_, **`mapping`**: _`chili.Mapper`_ = `None`) -> _`T`_\n\n`init_dataclass` function is instantiating dataclass of specified `type_name` and will hydrate the instance \nwith values passed in `value` dictionary. Each of the passed dictionary\'s keys must correspond to dataclass\'\nattributes in order to be properly interpreted. This rule can be broken if valid mapping is passed to the \nfunction.\n\nThis function support complex and nested hydration, which means if your dataclass aggregates other dataclasses \nor defines complex typing, `init_dataclass` function will respect your type annotations and will cast values \nto match the defined types. \n\nIf attributes in your dataclass do not specify the type value will be hydrated in to a newly created instance as is.\n### **`chili.asdict`**(**`value`**, **`mapping`**: _`chili.Mapper`_ = `None`) -> _`Dict[str, typing.Any]`_\n\n`asdict` is the opposite of `init_dataclass` function, it takes an instance of dataclass as argument, and\nextracts its members to a dictionary, so the returned data can be stored as json object or easily serialised \nto any other format. Additionally, `mapping` argument allows changing data representation on the fly.\n\n> Please note `Chili` is not a data validation library, although `Chili` performs some validation and casting\n> behind the scenes it does it only to ensure type consistency.\n\n',
    'author': 'Dawid Kraczkowski',
    'author_email': 'dawid.kraczkowski@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/kodemore/chili',
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
